from delvewheel._wheel_repair import WheelRepair
import sys
import pathlib

if __name__ == "__main__":

    args = sys.argv
    wheel_dir = args[1] if len(args) > 1 else "."
    wheel_dir = pathlib.Path(wheel_dir).absolute()
    print(f"Repairing wheels in {wheel_dir}")

    wheels = list(wheel_dir.glob("*.whl"))
    print(f"Found {len(wheels)} wheels")
    
    target_dir = pathlib.Path("./wheelhouse").absolute()
    for wheel in wheels:
        print(f"Repairing {wheel}")
        wr = WheelRepair(whl_path=wheel,
                         extract_dir=None,
                         add_dlls=None,
                         no_dlls=None,
                         ignore_in_wheel=True, 
                         verbose=0,
                         test=[])
        wr.show()
        wr.repair(target=target_dir,
                  no_mangles=set(),
                  no_mangle_all=False,
                  strip=False,
                  lib_sdir=".libs",
                  no_diagnostic=False)
                    


import toml
import os
import sys

if __name__ == "__main__":

    file = "pyproject.toml"
    if len(sys.argv) > 1:
        file = sys.argv[1]
        
    name = os.environ.get("PYPROJECT_NAME","llm-rs")

    try:
        data = toml.load(open(file,"r",encoding="utf-8"))
        data["project"]["name"] = name
        toml.dump(data,open("pyproject.toml.new","w",encoding="utf-8"))
        #Only replace if no error occured
        os.remove("pyproject.toml")
        os.rename("pyproject.toml.new","pyproject.toml")
        print(f"Changed projectname of file {file} to {name}!")
    except Exception as e:
        print(f"Could not change projectname of file {file} to {name}!")
        print(e)





from .config import QuantizationType,ContainerType,SessionConfig
import os
import pathlib
from .models import Mpt,GptNeoX,GptJ,Gpt2,Bloom,Llama
from .base_model import Model
import logging
from typing import Optional, List, Union,Type,Dict, Callable
import os
from enum import Enum, auto
from dataclasses import dataclass
import json
from blake3 import blake3
from huggingface_hub import snapshot_download
from huggingface_hub.utils import validate_repo_id, HFValidationError
from huggingface_hub import cached_assets_path

class QuantizationVersions(Enum):
    Not_Quantized=auto()
    V1=auto()
    V2=auto()

class KnownModels(Enum):
    GptNeoX = auto()
    Mpt = auto()
    GptJ = auto()
    Gpt2 = auto()
    Bloom = auto()
    Llama = auto()


_QUANTIZATION_TYPE_MAP = {
    "Q4_0": QuantizationType.Q4_0,
    "Q4_1": QuantizationType.Q4_1,
    "Q5_0": QuantizationType.Q5_0,
    "Q5_1": QuantizationType.Q5_1,
    "Q8_0": QuantizationType.Q8_0,
    "F16": QuantizationType.F16
}

_CONTAINER_TYPE_MAP = {
    "GGML": ContainerType.GGML,
    "GGJT": ContainerType.GGJT
}

_KNOWN_MODELS_MAP = {
    KnownModels.GptNeoX: GptNeoX,
    KnownModels.Mpt: Mpt,
    KnownModels.GptJ: GptJ,
    KnownModels.Gpt2: Gpt2,
    KnownModels.Bloom: Bloom,
    KnownModels.Llama: Llama
}

CURRENT_QUANTIZATION_VERSION = QuantizationVersions.V2

class PathType(Enum):
    DIR = auto()
    FILE = auto()
    REPO = auto()
    UNKNOWN = auto() 

def _get_path_type(path: Union[str,os.PathLike]) -> PathType:
    p = pathlib.Path(path)
    if p.is_file():
        return PathType.FILE
    elif p.is_dir():
        return PathType.DIR
    try:
        validate_repo_id(str(path))
        return PathType.REPO
    except HFValidationError:
        pass
    return PathType.UNKNOWN


@dataclass()
class ModelMetadata():
    """
    A dataclass to store metadata about a model.
    """
    model: KnownModels
    quantization: QuantizationType
    container: ContainerType
    quantization_version: QuantizationVersions=QuantizationVersions.Not_Quantized
    converter:str="llm-rs"
    hash:Optional[str]=None
    base_model:Optional[str]=None

    def add_hash(self,model_path:Union[str,os.PathLike]):
        h  = blake3(max_threads=blake3.AUTO)
        b  = bytearray(128_000_000)
        mv = memoryview(b)
        with open(model_path, 'rb', buffering=0) as f:
            for n in iter(lambda : f.readinto(mv), 0):
                h.update(mv[:n])
        self.hash=h.hexdigest()

    def serialize(self):
        return {
            "model": self.model.name,
            "quantization": repr(self.quantization).split(".")[-1],
            "quantization_version": self.quantization_version.name,
            "container": repr(self.container).split(".")[-1],
            "converter": self.converter,
            "hash": self.hash,
            "base_model": self.base_model
        }
    
    @staticmethod
    def deserialize(metadata_dict:Dict[str,str])->"ModelMetadata":
        return ModelMetadata(
            model = KnownModels[metadata_dict["model"]],
            quantization = _QUANTIZATION_TYPE_MAP[metadata_dict["quantization"]],
            quantization_version= QuantizationVersions[metadata_dict["quantization_version"]],
            container = _CONTAINER_TYPE_MAP[metadata_dict["container"]],
            converter = metadata_dict["converter"],
            hash = metadata_dict.get("hash"),
            base_model = metadata_dict.get("base_model")
        )


@dataclass
class AutoConfig():
    repo_type:Optional[str] = None
    model_type: Optional[str] = None

    @classmethod
    def from_pretrained(
        cls,
        model_path_or_repo_id: Union[str, os.PathLike],
        **kwargs,
    ) -> 'AutoConfig':
        path_type = _get_path_type(model_path_or_repo_id)
        if path_type == PathType.UNKNOWN:
            raise ValueError(
                f"Model path '{model_path_or_repo_id}' doesn't exist.")
        elif path_type == PathType.FILE:
            raise ValueError(
                f"Model path '{model_path_or_repo_id}' is a file. "
                "Please provide a directory or a repo id.")
        
        auto_config = AutoConfig()
        if path_type == PathType.DIR:
            cls._update_from_dir(str(model_path_or_repo_id), auto_config)
        elif path_type == PathType.REPO:
            cls._update_from_repo(str(model_path_or_repo_id), auto_config)

        return auto_config

    @classmethod
    def _update_from_repo(
        cls,
        repo_id: str,
        auto_config: 'AutoConfig',
    ) -> None:
        path = snapshot_download(repo_id=repo_id, allow_patterns='config.json')
        cls._update_from_dir(path, auto_config)

    @classmethod
    def _update_from_dir(cls, path: str, auto_config: 'AutoConfig') -> None:
        resolved_path = (pathlib.Path(path) / 'config.json').resolve()
        if resolved_path.is_file():
            cls._update_from_file(str(resolved_path), auto_config)
        else:
            raise ValueError(f"Config path '{resolved_path}' doesn't exist.")

    @classmethod
    def _update_from_file(cls, path: str, auto_config: 'AutoConfig') -> None:
        with open(path) as f:
            config = json.load(f)
        auto_config.model_type = config.get('model_type')
        auto_config.repo_type = config.get('repo_type')

class AutoModel():
    """
    Utility to load models, without having to specify the model type.
    """
    @classmethod
    def load_metadata(cls,model_file:Union[str,os.PathLike])->ModelMetadata:
        path = pathlib.Path(model_file)
        if not path.is_file():
            raise ValueError(f"Model file '{model_file}' is not a file!")
        if not path.exists():
            raise ValueError(f"Model file '{model_file}' does not exist!")

        metadata_file = path.with_suffix(".meta")
        if not metadata_file.exists():
            raise ValueError(f"Model file '{model_file}' does not have a metadata file '{metadata_file}'! If you want to autoload this model, please specify the model type.")
        
        metadata = ModelMetadata.deserialize(json.loads(metadata_file.read_text()))
        return metadata
    
    @classmethod
    def _infer_model_type(cls,model_file:Union[str,os.PathLike],known_model:Optional[KnownModels]=None)->Type[Model]:
        model_to_lookup = None
        if known_model:
            model_to_lookup = known_model
        else:
            metadata = cls.load_metadata(model_file)
            model_to_lookup = metadata.model

        if model_to_lookup in _KNOWN_MODELS_MAP:
            return _KNOWN_MODELS_MAP[model_to_lookup]
        else:
            raise ValueError(f"Unknown model type '{model_to_lookup}'")
            
        
    @classmethod
    def from_file(cls, path:Union[str,os.PathLike],
                  model_type: Optional[KnownModels] = None,
                  session_config:SessionConfig=SessionConfig(),
                  tokenizer_path_or_repo_id: Optional[Union[str,os.PathLike]]=None,
                  lora_paths:Optional[List[Union[str,os.PathLike]]]=None,
                  verbose:bool=False,
                  use_hf_tokenizer:bool=True)->Model:
        
        tokenizer = tokenizer_path_or_repo_id
        if use_hf_tokenizer and tokenizer is None:
            try:
                metadata = cls.load_metadata(path)
                tokenizer = metadata.base_model
            except Exception as e:
                logging.warning(f"Could not load metadata for model '{path}'!")

            if tokenizer is None or tokenizer == "":
                logging.warning(f"Model file '{path}' does not have a base_model specified in its metadata file but wants to use a huggingface-tokenizer! Please expilicitly specify a tokenizer via `tokenizer_path_or_repo_id` if you intend to use a huggingface-tokenizer.")

        model = cls._infer_model_type(path,model_type)
        return model(path,session_config,tokenizer_path_or_repo_id,lora_paths,verbose)
    
    @classmethod
    def from_pretrained(cls,
        model_path_or_repo_id: Union[str,os.PathLike],
        model_file: Optional[str] = None,
        model_type: Optional[KnownModels] = None,
        session_config:SessionConfig=SessionConfig(),
        tokenizer_path_or_repo_id: Optional[Union[str,os.PathLike]]=None,
        lora_paths:Optional[List[Union[str,os.PathLike]]]=None,
        verbose:bool=False,
        use_hf_tokenizer:bool=True,
        default_quantization:QuantizationType=QuantizationType.Q4_0,
        default_container:ContainerType=ContainerType.GGJT)->Model:

        path_type = _get_path_type(model_path_or_repo_id)

        if path_type == PathType.UNKNOWN:
            raise ValueError(f"Unknown path type for '{model_path_or_repo_id}'")
        elif path_type == PathType.FILE:
            return cls.from_file(model_path_or_repo_id,model_type,session_config,tokenizer_path_or_repo_id,lora_paths,verbose,use_hf_tokenizer)
        else:
            if path_type == PathType.REPO:

                try: 
                    config = AutoConfig.from_pretrained(
                        model_path_or_repo_id,
                    )
                except ValueError:
                    logging.warning("Could not find config.json in repo, assuming GGML model...")
                    config = AutoConfig(repo_type="GGML")
                       

                if config.repo_type != "GGML":
                    logging.warning("Found normal HuggingFace model, starting conversion...")
                    return cls.from_transformer(model_path_or_repo_id, session_config, tokenizer_path_or_repo_id, lora_paths, verbose, use_hf_tokenizer,default_quantization, default_container)
            
                resolved_path = cls._find_model_path_from_repo(str(model_path_or_repo_id),model_file)
                return cls.from_file(resolved_path,model_type,session_config,tokenizer_path_or_repo_id,lora_paths,verbose,use_hf_tokenizer)
            
            elif path_type == PathType.DIR:
                resolved_path = cls._find_model_path_from_dir(str(model_path_or_repo_id),model_file)
                return cls.from_file(resolved_path,model_type,session_config,tokenizer_path_or_repo_id,lora_paths,verbose,use_hf_tokenizer)
            
            else:
                raise ValueError(f"Unknown path type '{path_type}'")


    @classmethod
    def _find_model_path_from_dir(
        cls,
        directory: str,
        filename: Optional[str] = None,
    ) -> str:
        path = pathlib.Path(directory).resolve()
        if filename:
            file = (path / filename)
            if not file.is_file():
                raise ValueError(
                    f"Model file '{filename}' not found in '{path}'")
            return str(file)

        files = [
            f for f in path.iterdir()
            if f.is_file() and f.name.endswith('.bin')
        ]

        if len(files) == 0:
            raise ValueError(f"No model files found in '{path}'")
        elif len(files) > 1:
            raise ValueError(
                f"Multiple model files found in '{path}'! Please specify one of the following model files to load: '{','.join([f.name for f in files])}'"
            )

        return str(files[0].resolve())
    
    @classmethod
    def _find_model_path_from_repo(
        cls,
        repo_id: str,
        filename: Optional[str] = None,
    ) -> str:
        cache_directory = cached_assets_path("llm-rs",namespace=repo_id)
        #we dont want to download the whole repo, just the metadata
        directory = snapshot_download(repo_id=repo_id,
                                allow_patterns='*.meta',
                                local_dir=cache_directory)
        path = pathlib.Path(directory).resolve()
        files = [
            f for f in path.iterdir()
            if f.is_file() and f.name.endswith('.meta')
        ]

        if len(files) == 0 and filename is None:
            raise ValueError(f"No model files found in '{repo_id}'! Either specify the model file via the `model_file` parameter or make sure the repository contains a *.meta files for each model.")
        if len(files) > 1 and filename is None:
            raise ValueError(
                f"Multiple model files found in '{repo_id}'! Please specify one of the following model files to load: '{','.join([f.name.replace('.meta','.bin') for f in files])}' "
            )
        #if we have a filename, we can just download that file
        filename = filename if filename else files[0].name.replace('.meta','.bin')
        model_directory = snapshot_download(repo_id=repo_id,
                                allow_patterns=filename,
                                local_dir=cache_directory)
        
        return cls._find_model_path_from_dir(model_directory, filename=filename)
    
    @classmethod
    def from_transformer(cls,
        model_path_or_repo_id: Union[str,os.PathLike],
        session_config:SessionConfig=SessionConfig(),
        tokenizer_path_or_repo_id: Optional[Union[str,os.PathLike]]=None,
        lora_paths:Optional[List[Union[str,os.PathLike]]]=None,
        verbose:bool=False,
        use_hf_tokenizer:bool=True,
        default_quantization:QuantizationType=QuantizationType.Q4_0,
        default_container:ContainerType=ContainerType.GGJT):
        
        try:
            from .convert import AutoConverter
            from .convert.auto_converter import get_name_from_config
            from transformers import AutoConfig
        except ImportError:
            raise ImportError("Model conversion needs additional dependencies. Please install 'llm-rs[convert]' to use this functionality!")
        

        config = AutoConfig.from_pretrained(model_path_or_repo_id,trust_remote_code=True)
        model_name = get_name_from_config(config)
        export_path = cached_assets_path("llm-rs",namespace=model_name)
        converted_model = AutoConverter.convert(model_path_or_repo_id,export_path)
        if default_quantization != QuantizationType.F16:
            converted_model = AutoQuantizer.quantize(converted_model,quantization=default_quantization,container=default_container)
        return cls.from_file(converted_model,None,session_config,tokenizer_path_or_repo_id,lora_paths,verbose,use_hf_tokenizer)
    
# Hack to make the quantization type enum hashable
_APPENDIX_MAP = {
    QuantizationType.Q4_0.__repr__(): "q4_0",
    QuantizationType.Q4_1.__repr__(): "q4_1",
    QuantizationType.Q5_0.__repr__(): "q5_0",
    QuantizationType.Q5_1.__repr__(): "q5_1",
    QuantizationType.Q8_0.__repr__(): "q8_0",
}

class AutoQuantizer():
    """
    Utility to quantize models, without having to specify the model type.
    """
    @staticmethod
    def quantize(
        model_file:Union[str,os.PathLike],
        target_path:Optional[Union[str,os.PathLike]]=None,
        quantization:QuantizationType=QuantizationType.Q4_0,
        container:ContainerType=ContainerType.GGJT,
        callback:Optional[Callable[[str],None]]=None
        )->Union[str,os.PathLike]:
        metadata=AutoModel.load_metadata(model_file)
        if metadata.quantization != QuantizationType.F16:
            raise ValueError(f"Model '{model_file}' is already quantized to '{metadata.quantization}'")

        model_type = AutoModel._infer_model_type(model_file)

        if target_path is None:
            target_path = os.path.dirname(model_file)
            
        def build_target_name()->str:
            output_path = pathlib.Path(target_path)
            if output_path.is_file():
                return str(output_path)
            else:
                output_path.mkdir(parents=True,exist_ok=True)
                model_path = pathlib.Path(model_file)
                appendix = ""

                if quantization.__repr__() in _APPENDIX_MAP:
                    appendix += f"-{_APPENDIX_MAP[quantization.__repr__()]}"

                if container == ContainerType.GGJT:
                    appendix += "-ggjt"
            
                filename = model_path.stem.replace("-f16","") + appendix + model_path.suffix
                return str(output_path / filename)

        target_file = build_target_name()
        if pathlib.Path(target_file).exists():
            logging.warning(f"Target file '{target_file}' already exists, skipping quantization")
            return target_file
        
        logging.info(f"Quantizing model '{model_file}' to '{target_file}'")
        model_type.quantize(str(model_file),target_file,quantization,container,callback=callback)

        metadata_file = pathlib.Path(target_file).with_suffix(".meta")
        quantized_metadata = ModelMetadata(model=metadata.model,quantization=quantization,container=container,quantization_version=CURRENT_QUANTIZATION_VERSION,base_model=metadata.base_model)
        quantized_metadata.add_hash(target_file)
        logging.info(f"Writing metadata file '{metadata_file}'")
        metadata_file.write_text(json.dumps(quantized_metadata.serialize(),indent=4))
        logging.info(f"Finished quantizing model '{model_file}' to '{target_file}'")
        return target_file





try:
    from .llm_rs import *
except ImportError as e:
    print("DLLs were not boundled with this package. Trying to locate them...")
    import os
    import platform

    #Try to locate CUDA_PATH environment variable
    cuda_path = os.environ.get("CUDA_PATH",None)
    if cuda_path:
        print(f"Found CUDA_PATH environment variable: {cuda_path}")
        if platform.system() == "Windows":
            cuda_path = os.path.join(cuda_path,"bin")
        else:
            cuda_path = os.path.join(cuda_path,"lib64")

        print(f"Adding {cuda_path} to DLL search path...")
        os.add_dll_directory(cuda_path)

    try:
        from .llm_rs import *
    except ImportError as inner_e:
        raise  ImportError("Could not locate DLLs. Please check the documentation for more information.")
   


from .config import GenerationConfig, SessionConfig, Precision, ContainerType, QuantizationType
from .models import Llama, GptJ, Gpt2, Bloom, GptNeoX, Mpt
from .auto import AutoModel, ModelMetadata, KnownModels, AutoQuantizer

__doc__ = llm_rs.__doc__
if hasattr(llm_rs, "__all__"):
    __all__ = llm_rs.__all__


from haystack.nodes import PromptNode, PromptModel
from llm_rs.haystack import RustformersInvocationLayer

model = PromptModel("rustformers/open-llama-ggml",
                    max_length=1024,
                    invocation_layer_class=RustformersInvocationLayer,
                    model_kwargs={"model_file":"open_llama_3b-q5_1-ggjt.bin"})

pn = PromptNode(
    model,
    max_length=1024
)

pn("Life is",stream=True)




from llm_rs.langchain import RustformersLLM
from langchain import PromptTemplate
from langchain.chains import LLMChain
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler

template="""Below is an instruction that describes a task. Write a response that appropriately completes the request.
### Instruction:
{instruction}
### Response:
Answer:"""

prompt = PromptTemplate(input_variables=["instruction"],template=template,)

llm = RustformersLLM(model_path_or_repo_id="rustformers/mpt-7b-ggml",model_file="mpt-7b-instruct-q5_1-ggjt.bin",callbacks=[StreamingStdOutCallbackHandler()])

chain = LLMChain(llm=llm, prompt=prompt)

chain.run("Write a short post congratulating rustformers on their new release of their langchain integration.")


try:
    from haystack.nodes import PromptModelInvocationLayer
    from haystack.nodes.prompt.invocation_layer import DefaultTokenStreamingHandler
except ImportError:
     raise ImportError(
        'To use the llm_rs.haystack module, please install llm-rs with the additional "haystack" dependencies via: pip install llm-rs[haystack]')

import os
from typing import Dict, List, Union, Type, Optional
from ..auto import AutoModel,KnownModels
from ..config import QuantizationType,ContainerType,SessionConfig,GenerationConfig
from ..base_model import Model
import logging 

logger = logging.getLogger(__name__)

class RustformersInvocationLayer(PromptModelInvocationLayer):
    def __init__(self, model_name_or_path: Union[str,os.PathLike], 
        max_length: int = 512,
        model_file: Optional[str] = None,
        model_type: Optional[KnownModels] = None,
        session_config:SessionConfig=SessionConfig(),
        tokenizer_path_or_repo_id: Optional[Union[str,os.PathLike]]=None,
        lora_paths:Optional[List[Union[str,os.PathLike]]]=None,
        verbose:bool=False,
        use_hf_tokenizer:bool=True,
        default_quantization:QuantizationType=QuantizationType.Q4_0,
        default_container:ContainerType=ContainerType.GGJT,
        **kwargs):

        """
        Creates a new RustformersInvocationLayer instance.

        :param model_name_or_path: The name or path of the underlying model.
        :param kwargs: See `AutoModel.from_pretrained`.
        """
        if model_name_or_path is None or len(model_name_or_path) == 0:
            raise ValueError("model_name_or_path cannot be None or empty string")

        self.model_name_or_path = model_name_or_path
        self.max_length = max_length 
        self.model:Model = AutoModel.from_pretrained(model_path_or_repo_id = model_name_or_path,
            model_file=model_file,
            model_type=model_type,
            session_config=session_config,
            tokenizer_path_or_repo_id=tokenizer_path_or_repo_id,
            lora_paths=lora_paths,
            verbose=verbose,
            use_hf_tokenizer=use_hf_tokenizer,
            default_quantization=default_quantization,
            default_container=default_container)
        

    def _ensure_token_limit(self, prompt: Union[str, List[Dict[str, str]]]) -> Union[str, List[Dict[str, str]]]:
        """Ensure that length of the prompt and answer is within the maximum token length of the PromptModel.

        :param prompt: Prompt text to be sent to the generative model.
        """
        if not isinstance(prompt, str):
            raise ValueError(f"Prompt must be of type str but got {type(prompt)}")
        
        context_length = self.model.config.context_length
        tokenized_prompt = self.model.tokenize(prompt)
        if len(tokenized_prompt) + self.max_length > context_length:
            logger.warning(
            "The prompt has been truncated from %s tokens to %s tokens so that the prompt length and "
            "answer length (%s tokens) fit within the max token limit (%s tokens). "
            "Shorten the prompt to prevent it from being cut off",
            len(tokenized_prompt),
            max(0, context_length -  self.max_length),
            self.max_length,
            context_length,
            )
            return self.model.decode(tokenized_prompt[:max(0, context_length -  self.max_length)])

        return prompt

    def invoke(self, *args, **kwargs):
        """
        It takes a prompt and returns a list of generated text using the underlying model.
        :return: A list of generated text.
        """

        prompt = kwargs.pop("prompt",None)

        if not prompt:
            raise ValueError("`prompt` cannot be None")
    
        generation_config = kwargs.pop("generation_config", GenerationConfig())
        generation_config.max_new_tokens = self.max_length

        stream = kwargs.pop("stream", False)
        stream_handler = kwargs.get("stream_handler", DefaultTokenStreamingHandler())

        generated_text = []
        for token in self.model.stream(prompt=prompt,generation_config=generation_config):
            generated_text.append(token)
            if stream:
                stream_handler(token)

        return generated_text

    def supports(cls, model_name_or_path: str, **kwargs) -> bool:
        """
        Checks if the given model is supported by this invocation layer.

        :param model_name_or_path: The name or path of the model.
        :param kwargs: Additional keyword arguments passed to the underlying model which might be used to determine
        if the model is supported.
        :return: True if this invocation layer supports the model, False otherwise.
        """
        #I guess there is not much to validate here ¯\_(ツ)_/¯
        return model_name_or_path is not None and len(model_name_or_path) > 0